from logging import config
import os
from typing import Any, Dict, Optional, Callable, List
from acex_devkit.configdiffer import Diff
from acex_devkit.configdiffer.command import Command, Context
from acex.plugins.neds.core import RendererBase
from acex_devkit.models.composed_configuration import ComposedConfiguration
from acex_devkit.models.logging import LoggingSeverity
from pathlib import Path

from jinja2 import Environment, FileSystemLoader, StrictUndefined
from .filters import cidr_to_addrmask
from .hardware_models import match_hardware_model

from .device_types.resolver import get_model


class GeneratorRegistry:
    def __init__(self):
        self._patterns: list[tuple[tuple, Callable]] = []

    def register(self, pattern: tuple, generator: Callable):
        self._patterns.append((pattern, generator))

    def resolve(self, path: tuple):
        for pattern, generator in self._patterns:
            if self._match(path, pattern):
                return generator
        return None

    def _match(self, path, pattern):
        if len(path) < len(pattern):
            return False

        for p, pat in zip(path, pattern):
            if pat != "*" and p != pat:
                return False
        return True


class CiscoIOSCLIRenderer(RendererBase):

    @property
    def _model_directory(self) -> str:
        """Return the directory path for Cisco models."""
        package_dir = os.path.dirname(__file__)
        return os.path.join(package_dir, "device_types/models")

    def _load_template_file(self, asset) -> str:
        """Load a Jinja2 template file."""
        # handle cluster and single assets.
        template_name = "template.j2"

        path = Path(__file__).parent
        env = Environment(
            loader=FileSystemLoader(path), undefined=StrictUndefined
        )  # StrictUndefined to catch undefined variables, testing

        env.filters["cidr_to_addrmask"] = cidr_to_addrmask
        template = env.get_template(template_name)
        return template


    def _register(self):
        """
        Registers patterns/generators to registry.

        for specific command generators, register them each and
        map them to a pattern in the generator-registry
        """
        self.registry.register(
            ("system", "config"), self._generate_system_config_commands
        )
        self.registry.register(
            ("interfaces", "*"), self._generate_interface_config_commands
        )

    def flatten(self, commands):
        output = []
        current_ctx = ()

        for cmd in commands:
            target_ctx = cmd.context.path

            # Om vi byter context
            if target_ctx != current_ctx:

                # Gå upp tills common prefix
                while not target_ctx[: len(current_ctx)] == current_ctx:
                    output.append("exit")
                    current_ctx = current_ctx[:-1]

                # Gå ner i nytt context
                for part in target_ctx[len(current_ctx) :]:
                    output.append(part)
                    current_ctx += (part,)

            output.append(f" {cmd.command}")

        # Stäng alla contexts
        while current_ctx:
            output.append("exit")
            current_ctx = current_ctx[:-1]

        return output

    # Render config patches from diff below, move to better place laterz
    def render_patch(self, diff: Diff, node_instance: "NodeInstance"):
        """
        Render specific commands for patching based on a diff.
        """
        # Create GeneratorRegistry and register patterns
        self.registry = GeneratorRegistry()
        self._register()

        commands = []
        for change in diff.get_all_changes():
            generator = self.registry.resolve(tuple(change.path))
            if generator is None:
                continue
            cmds = generator(change, node_instance)
            commands.extend(cmds)

        # TODO: Ordering

        # TODO: render context

        flat_commands = self.flatten(commands)
        cli_text = "!\r\n"
        for line in flat_commands:
            cli_text += line + "\r\n"
        cli_text += "!"
        return cli_text

    def render(self, configuration: ComposedConfiguration, asset) -> Any:
        """Render the configuration model for Cisco IOS CLI devices."""

        """ TODO:
        - Interface preprocess
        - os och os version? ta från första noden? 

        """

        if isinstance(configuration, ComposedConfiguration):
            configuration = configuration.model_dump(mode="json")
        else:
            raise ValueError(
                f"Configuration must be a ComposedConfiguration instance. Not {type(configuration)}"
            )

        # Give the NED a chance to pre-process the config before rendering
        processed_config = self.pre_process(configuration, asset)

        # Render template and return payload
        template = self._load_template_file(asset)
        return template.render(configuration=processed_config)

    def _generate_system_config_commands(
        self, component_change, node_instance
    ) -> List[Command]:
        """
        Generate config related to system config.
        """
        ctx = Context(path=[])  # tom eftersom dessa körs i rootnivå
        commands = []

        for attr in component_change.changed_attributes:
            if attr.attribute_name == "hostname":
                if component_change.op in ("add", "change"):
                    cmd_txt = f"hostname {attr.after.value}"
                elif component_change.op == "remove":
                    cmd_txt = "no hostname"

                commands.append(Command(context=ctx, command=cmd_txt))
        return commands

    def _generate_interface_config_commands(
        self, component_change, node_instance
    ) -> List[Command]:
        """
        Generate config related to interface config.
        """
        ctx = Context(path=component_change.path)
        commands = []
        cmd = Command(context=ctx, command="desc hyvvää hirvi")

        commands.append(cmd)
        return commands

    def _ssh_interface(self, config):
        # ip ssh source-interface {'pointer': 'interfaces.vlan123_svi', 'metadata': {'type': 'str', 'value_source': 'reference'}}
        ssh_config = config.get("system", {}).get("ssh")
        ssh_raw_interface = (
            ssh_config.get("config", {}).get("source_interface") if ssh_config else None
        )
        if ssh_raw_interface and isinstance(ssh_raw_interface, dict):
            ref_path = ssh_raw_interface.get("pointer")
            if ref_path:
                ref_name = ref_path.split(".")[1]
                intf = config.get("interfaces", {}).get(ref_name)
                if intf:
                    vlan_id = intf.get("vlan_id")
                    if vlan_id is not None:
                        ssh_interface = f"Vlan{vlan_id.get('value')}"
                        # replace current source interface with formatted one for template use
                        ssh_config["config"]["source_interface"] = ssh_interface

        return config

    def _logging_trap_severity(self, config):
        logging_trap = (
            config.get("system", {})
            .get("logging", {})
            .get("config", {})
            .get("augments")
            or {}
        ).get("cisco.trap_logging")
        if logging_trap and logging_trap.get("severity"):
            # Translate severity to Cisco IOS format
            raw_severity = LoggingSeverity(logging_trap["severity"].get("value"))
            if "NOTICE" in raw_severity.value:
                logging_trap["severity"]["value"] = "informational"
            else:
                logging_trap["severity"]["value"] = raw_severity.value.lower()

        return config

    def _add_vrf_to_intefaces(self, config):
        """
        Loops all network_instances and add vrf definition to
        referenced interfaces
        """
        vrfs = config["network_instances"]
        for vrf_name, vrf in vrfs.items():
            if vrf["name"]["value"] == "global":
                ...
            else:
                for _, interface in vrf["interfaces"].items():
                    ref_path = interface['pointer']
                    if isinstance(ref_path, str) and ref_path:
                        intf = config["interfaces"][ref_path.split(".")[1]]
                        intf["vrf"] = vrf["name"]["value"]

    def pre_process(self, configuration, asset) -> Dict[str, Any]:
        """Pre-process the configuration model before rendering j2."""

        # Render physical interface config
        self._add_vrf_to_intefaces(configuration)
        self._render_frontpanel_ports(configuration, asset)
        self._ssh_interface(configuration)
        self._logging_trap_severity(configuration)
        # self.lacp_load_balancing(configuration)

        if hasattr(asset, "assets"):
            os_version = asset.assets[0].os_version
        else:
            os_version = asset.os_version

        configuration["asset"] = {"version": os_version}
        return configuration

    def _render_frontpanel_ports(self, configuration, asset):
        """
        Entry method for rendering frontpanel interfaces. 
        If asset.type is cluster or single asset, this 
        method resolves and loops accordingly. 
        """
        if asset.type == "asset":
            model_data = get_model(asset.hardware_model, self._model_directory)
            port_slots = self._render_frontpanel_port_slots(0, configuration, model_data)
            configuration = self._update_frontpanel_ports(configuration, port_slots)
        elif asset.type == "asset_cluster":
            all_slots = []
            for stack_index, asset in enumerate(asset.assets):
                model_data = get_model(asset.hardware_model, self._model_directory)
                all_slots.extend(self._render_frontpanel_port_slots(stack_index, configuration, model_data))

            configuration = self._update_frontpanel_ports(configuration, all_slots)



    def _render_frontpanel_port_slots(self, stack_index, configuration, model_data):
        """
        Build a list of "interface slots". These slots are to be filled with config
        from configuration, such as speed, before passed to template rendering.
        """
        intf_slots = []

        interface_pattern = model_data["name_pattern"]
        prefix_map = model_data["prefix_map"]
        idx_start = model_data["interface_index_start"]
        midx_start = model_data["module_index_start"]
        sidx_start = model_data["stack_index_start"]

        for interface_slot in model_data["interfaces"]:

            # Get prefix based on configured speed
            speed = self._get_configured_speed(configuration, interface_slot)
            prefix = prefix_map.get(speed)

            # Compile full interface name
            ifname = self._compile_interface_name(
                prefix=prefix,
                stack_index=stack_index,
                interface_slot=interface_slot,
                model_data=model_data
            )

            intf_slots.append({
                "name": ifname,
                "index": interface_slot["index"],
                "module_index": interface_slot["module_index"],
                "stack_index": stack_index
                })

        return intf_slots


    def _update_frontpanel_ports(self, configuration, slots):
        """
        Fetch corresponding configuration for each frontpanel port slot.
        """
        interfaces_without_slots = []

        for k,v in configuration["interfaces"].items():
            if v["type"] == "ethernetCsmacd":
                idx = v["index"]["value"]
                midx = v["module_index"]["value"]
                sidx = v["stack_index"]["value"]
                slot = self._get_slot(idx, midx, sidx, slots)

                if slot is None:
                    interfaces_without_slots.append(k)
                else:
                    v["name"] = slot["name"]

        for interface in interfaces_without_slots:
            configuration["interfaces"].pop(interface)

        return configuration


    def _get_slot(self, index, module_index, stack_index, slots):
        """ Extracts a slot definition or None"""
        for slot in slots:
            if slot["index"] == index and slot["module_index"] == module_index and slot["stack_index"] == stack_index:
                return slot
        return None


    def _compile_interface_name(self, *, prefix: str, stack_index:int, interface_slot:dict, model_data: dict):
        """
        Compiles full interface name from device_types/models file
        """

        pattern = model_data["name_pattern"]
        idx_offset = model_data["interface_index_start"]
        midx_offset = model_data["module_index_start"]
        sidx_offset = model_data["stack_index_start"]

        name = pattern.format_map({
            "prefix": prefix,
            "stack_index": stack_index + sidx_offset,
            "module_index":  interface_slot["module_index"] + midx_offset,
            "index":  interface_slot["index"] + idx_offset
        })

        return name
        

    def _get_configured_speed(self, configuration: dict, interface_slot: dict):
        """ extract configured interface speed from config map """

        for k,v in configuration["interfaces"].items():
            configured_speed = v.get('speed', {}).get('value', 1000000) 

        return configured_speed


"""Cisco IOS / IOS-XE normalizer — vendor-specific rules only."""

import re

from acex_devkit.normalizer import BaseNormalizer, LineRule, BlockRule, RewriteRule


class CiscoIOSNormalizer(BaseNormalizer):

    line_rules = [
        LineRule("header_building",   re.compile(r"^Building configuration\.\.\.")),
        LineRule("header_current",    re.compile(r"^Current configuration\s*:\s*\d+\s*bytes")),
        LineRule("last_change",       re.compile(r"^!\s*Last configuration change")),
        LineRule("nvram_updated",     re.compile(r"^!\s*NVRAM config last updated")),
        LineRule("no_change",         re.compile(r"^!\s*No configuration change")),
        LineRule("ntp_clock_period",  re.compile(r"^\s*ntp clock-period\s+\d+")),
        LineRule("license_udi",       re.compile(r"^\s*license udi\s+")),
        LineRule("license_boot",      re.compile(r"^\s*license boot level\s+")),
        LineRule("snmp_engineid",     re.compile(r"^\s*snmp-server engineID\s+")),
        LineRule("cts_sxp_dynamic",   re.compile(r"^\s*cts sxp connection peer\s+")),
        LineRule("end_marker",        re.compile(r"^end\s*$")),
        LineRule("version_banner",    re.compile(r"^version\s+\d+\.\d+\s*$")),
        LineRule("boot_marker",       re.compile(r"^\s*boot-start-marker\s*$")),
        LineRule("boot_end_marker",   re.compile(r"^\s*boot-end-marker\s*$")),
    ]

    block_rules = [
        BlockRule("pki_cert_chain",       re.compile(r"^crypto pki certificate chain\s+")),
        BlockRule("pki_trustpoint_self",  re.compile(r"^crypto pki trustpoint\s+TP-self-signed-")),
        BlockRule("pki_certificate_self", re.compile(r"^crypto pki certificate\s+self-signed")),
    ]

    rewrite_rules = [
        RewriteRule(
            "enable_secret",
            re.compile(r"^(\s*enable secret\s+\d+)\s+\S+.*$"),
            r"\1 <REDACTED>",
        ),
        RewriteRule(
            "enable_password",
            re.compile(r"^(\s*enable password(?:\s+\d+)?)\s+\S+.*$"),
            r"\1 <REDACTED>",
        ),
        RewriteRule(
            "username_secret",
            re.compile(r"^(\s*username\s+\S+(?:\s+privilege\s+\d+)?\s+secret\s+\d+)\s+\S+.*$"),
            r"\1 <REDACTED>",
        ),
        RewriteRule(
            "username_password",
            re.compile(r"^(\s*username\s+\S+(?:\s+privilege\s+\d+)?\s+password(?:\s+\d+)?)\s+\S+.*$"),
            r"\1 <REDACTED>",
        ),
        RewriteRule(
            "snmp_community",
            re.compile(r"^(\s*snmp-server community)\s+\S+(\s+.*)?$"),
            r"\1 <REDACTED>\2",
        ),
        RewriteRule(
            "aaa_shared_key",
            re.compile(r"^(\s*key(?:\s+\d+)?)\s+\S+.*$"),
            r"\1 <REDACTED>",
        ),
        RewriteRule(
            "ike_psk",
            re.compile(r"^(\s*pre-shared-key(?:\s+address\s+\S+)?)\s+key\s+\S+.*$"),
            r"\1 key <REDACTED>",
        ),
        RewriteRule(
            "ospf_md5_key",
            re.compile(r"^(\s*ip ospf message-digest-key\s+\d+\s+md5(?:\s+\d+)?)\s+\S+.*$"),
            r"\1 <REDACTED>",
        ),
        RewriteRule(
            "bgp_password",
            re.compile(r"^(\s*neighbor\s+\S+\s+password(?:\s+\d+)?)\s+\S+.*$"),
            r"\1 <REDACTED>",
        ),
    ]




import asyncio
from contextlib import asynccontextmanager
from contextvars import ContextVar
from typing import Optional
from acex_devkit.models.composed_configuration import ComposedConfiguration
from acex_devkit.models.node_response import NodeListItem
from acex_devkit.models.management_connection import ManagementConnection
from scrapli.driver.core import AsyncIOSXEDriver
from scrapli.exceptions import ScrapliAuthenticationFailed, ScrapliTimeout

from acex_devkit.drivers import NetworkElementDriver, TransportBase
from acex_devkit.exceptions import AuthenticationFailed, ConnectionTimeout
from acex_devkit.configdiffer import Diff

from .renderer import CiscoIOSCLIRenderer
from .parser import CiscoIOSCLIParser
from .normalizer import CiscoIOSNormalizer

# Legacy KEX/ciphers for older IOS devices that don't support modern algorithms
_LEGACY_ASYNCSSH_OPTIONS = {
    "kex_algs": [
        "diffie-hellman-group-exchange-sha256",
        "diffie-hellman-group-exchange-sha1",
        "diffie-hellman-group14-sha256",
        "diffie-hellman-group14-sha1",
        "diffie-hellman-group1-sha1",
    ],
}


_session_conn: ContextVar[Optional[AsyncIOSXEDriver]] = ContextVar("_session_conn", default=None)


class CiscoIOSTransport(TransportBase):

    async def _open_connection(self, connection: ManagementConnection, **kwargs) -> AsyncIOSXEDriver:
        username = kwargs.get("username")
        password = kwargs.get("password")
        if not username or not password:
            raise ValueError("Credentials required: username and password must be provided")
        driver = AsyncIOSXEDriver(
            host=connection.target_ip,
            auth_username=username,
            auth_password=password,
            auth_secondary=kwargs.get("enable_password") or password,
            auth_strict_key=False,
            port=22,
            timeout_socket=30,
            timeout_ops=120,
            transport="asyncssh",
            transport_options={"asyncssh": _LEGACY_ASYNCSSH_OPTIONS},
        )
        try:
            await driver.open()
        except ScrapliAuthenticationFailed as e:
            if isinstance(e.__cause__, asyncio.TimeoutError):
                raise ConnectionTimeout("Connection timed out") from e
            raise AuthenticationFailed("Authentication failed") from e
        except ScrapliTimeout as e:
            raise ConnectionTimeout("Connection timed out") from e
        return driver

    @asynccontextmanager
    async def session(self, connection: ManagementConnection, **kwargs):
        """Hold one SSH session open for the duration of the block."""
        conn = await self._open_connection(connection, **kwargs)
        token = _session_conn.set(conn)
        try:
            yield self
        finally:
            _session_conn.reset(token)
            try:
                await conn.close()
            except Exception:
                pass

    @asynccontextmanager
    async def _conn(self, connection: ManagementConnection, **kwargs):
        """Yield active session conn or open a one-shot, closing what we own."""
        conn = _session_conn.get()
        if conn is not None:
            yield conn
            return
        conn = await self._open_connection(connection, **kwargs)
        try:
            yield conn
        finally:
            try:
                await conn.close()
            except Exception:
                pass

    async def get_config(self, node: NodeListItem, connection: ManagementConnection, **kwargs) -> str:
        async with self._conn(connection, **kwargs) as conn:
            response = await conn.send_command("show running-config", timeout_ops=120)
            return response.result

    async def send_config(self, node: NodeListItem, connection: ManagementConnection, commands: list[str], **kwargs) -> str:
        async with self._conn(connection, **kwargs) as conn:
            response = await conn.send_configs(commands)
            return response.result

    async def execute(self, node: NodeListItem, connection: ManagementConnection, commands: list[str], **kwargs) -> list[str]:
        async with self._conn(connection, **kwargs) as conn:
            responses = await conn.send_commands(commands, timeout_ops=120)
            return [r.result for r in responses]

    async def get_lldp_neighbors(self, node: NodeListItem, connection: ManagementConnection, **kwargs) -> list[dict]:
        async with self._conn(connection, **kwargs) as conn:
            neighbors = []
            try:
                response = await asyncio.wait_for(conn.send_command("show lldp neighbors detail"), timeout=10)
                if "%" not in response.result[:40]:
                    neighbors.extend(self._parse_lldp_detail(response.result))
            except (asyncio.TimeoutError, ScrapliTimeout, ConnectionTimeout, Exception):
                # Connection state is unknown after timeout — skip CDP on this session
                return neighbors
            try:
                response = await asyncio.wait_for(conn.send_command("show cdp neighbors detail"), timeout=10)
                if "%" not in response.result[:40]:
                    seen = {(n["local_interface"], n["remote_device"]) for n in neighbors}
                    for entry in self._parse_cdp_detail(response.result):
                        if (entry["local_interface"], entry["remote_device"]) not in seen:
                            neighbors.append(entry)
            except (asyncio.TimeoutError, ScrapliTimeout, ConnectionTimeout, Exception):
                pass
            return neighbors

    @staticmethod
    def _parse_lldp_detail(raw: str) -> list[dict]:
        import re
        entries = re.split(r'-{20,}', raw)
        neighbors = []
        for entry in entries:
            local = re.search(r'Local Intf:\s*(\S+)', entry)
            sys_name = re.search(r'System Name:\s*(\S+)', entry)
            port_id = re.search(r'Port id:\s*(\S+)', entry)
            if local and sys_name:
                neighbors.append({
                    "local_interface": local.group(1),
                    "remote_device": sys_name.group(1),
                    "remote_interface": port_id.group(1) if port_id else "",
                    "discovery_protocol": "lldp",
                })
        return neighbors

    @staticmethod
    def _parse_cdp_detail(raw: str) -> list[dict]:
        import re
        entries = re.split(r'-{20,}', raw)
        neighbors = []
        for entry in entries:
            device = re.search(r'Device ID:\s*(\S+)', entry)
            local = re.search(r'Interface:\s*(\S+?),', entry)
            remote = re.search(r'Port ID.*?:\s*(\S+)', entry)
            if device and local:
                neighbors.append({
                    "local_interface": local.group(1),
                    "remote_device": device.group(1).rstrip('.'),
                    "remote_interface": remote.group(1) if remote else "",
                    "discovery_protocol": "cdp",
                })
        return neighbors


class CiscoIOSCLIDriver(NetworkElementDriver):
    """Cisco IOS CLI driver."""

    renderer_class = CiscoIOSCLIRenderer
    transport_class = CiscoIOSTransport
    parser_class = CiscoIOSCLIParser
    normalizer_class = CiscoIOSNormalizer

    def render(self, configuration: ComposedConfiguration, asset):
        return self.renderer.render(configuration, asset)

    def parse(self, configuration: str) -> ComposedConfiguration:
        return self.parser.parse(configuration)

    def render_patch(self, diff: Diff, node_instance: "NodeInstance"):
        return self.renderer.render_patch(diff, node_instance)

    async def apply_patch(self, diff: Diff, node_instance, node: NodeListItem, connection: ManagementConnection, **kwargs):
        commands = self.render_patch(diff, node_instance=node_instance)
        commands = [c.lstrip() for c in commands.splitlines() if c.strip() != "!"]
        return await self.transport.send_config(node, connection, commands, **kwargs)



# from acex.models.composed_configuration import ComposedConfiguration, EthernetCsmacdInterface, L3IpvlanInterface, SoftwareLoopbackInterface, Ieee8023adLagInterface
from acex_devkit.models.composed_configuration import (
    AuthorizedKey,
    AuthorizedKeyAlgorithms,
    ComposedConfiguration,
    EthernetCsmacdInterface,
    L3IpvlanInterface,
    SoftwareLoopbackInterface,
    Ieee8023adLagInterface,
    SshServer,
    NtpServer,
    SystemConfig,
    Snmp,
    SnmpConfig,
    SnmpCommunity,
    SnmpUser,
    SnmpSecurityLevel,
    SnmpView,
    SnmpViewAttributes,
    SnmpServer,
    TrapEvent,
    ReferenceTo,
)
from ntc_templates.parse import parse_output
import os


def map_enabled(enabled_value: str) -> bool:
    """
    Map TextFSM 'ENABLED' field to boolean.
    True = interface is up (no shutdown or empty)
    False = interface is shutdown
    """
    return enabled_value != "shutdown"


def expand_vlans(vlan_str: str) -> list[int]:
    """
    Expand VLAN string representation to a list of integers.
    E.g., "1-3,5" -> [1, 2, 3, 5]
    """
    vlans = set()
    for part in vlan_str.split(","):
        if "-" in part:
            start, end = map(int, part.split("-"))
            vlans.update(range(start, end + 1))
        else:
            vlans.add(int(part))
    return sorted(vlans)


class CiscoIOSCLIParser:
    def __init__(self):
        self.running_config = None
        self._parsed_config = ComposedConfiguration()

    @property
    def parsed_config(self):
        """Return the model class for the parser."""
        return self._parsed_config

    @property
    def custom_templates_dir(self) -> str:
        """Return the directory path for custom TextFSM templates."""
        package_dir = os.path.dirname(__file__)
        return os.path.join(package_dir, "custom_textfsm_templates")

    @property
    def platform(self) -> str:
        """Return the platform name for the parser."""
        return "cisco_ios"

    # Mostly used to remove metadata key, which contains non-serializable data and is not needed in the final output, but can be used to remove any key if needed in the future
    def removekey(self, d, key):
        if hasattr(d, "model_dump"):  # Pydantic v2
            r = d.model_dump()
            # print('model_dump: ', r)
        elif hasattr(d, "dict"):  # Pydantic v1
            r = d.dict()
            # print('dict: ', r)
        else:
            r = dict(d)  # Fallback for regular dicts or other types
            # print('dict fallback: ', r)

        def _remove(obj):
            if isinstance(obj, dict):
                # print('removing key from dict: ', obj)
                return {k: _remove(v) for k, v in obj.items() if k != key}
            elif isinstance(
                obj, list
            ):  # supporting in case we use lists of dicts in the future
                # print('removing key from list: ', obj)
                return [_remove(item) for item in obj]
            # print('returning obj: ', obj)
            return obj

        return _remove(r)

    def parse(self, configuration: str) -> dict:
        """Parse the Cisco IOS CLI configuration content."""
        self.running_config = configuration
        self.parse_system_settings()
        self.parse_interfaces()
        self.parse_l3_interfaces()
        self.parse_ntp()
        self.parse_ssh()
        self.parse_snmp()
        self.parse_snmp_servers()
        self.parse_snmp_views()
        self.parse_snmp_communities()
        self.parse_snmp_users()

        return self._parsed_config

    def parse_l3_interfaces(self):
        """Parse L3 interfaces."""
        command = "show running svi interfaces"

        parsed_data = parse_output(
            platform=self.platform,
            template_dir=self.custom_templates_dir,
            command=command,
            data=self.running_config,
        )

        for intf in parsed_data:
            intf["enabled"] = map_enabled(intf.get("ENABLED", ""))
            intf["description"] = intf.get("description") or None
            intf["ipv4"] = intf.get("ipv4_address") or None
            # intf['ipv6_address'] = intf.get('ipv6_address') or None
            intf["vlan_id"] = int(intf["name"].replace("Vlan", ""))

        interfaces_dict = {
            intf["name"]: self.removekey(
                L3IpvlanInterface(index=index, **intf), "metadata"
            )
            for index, intf in enumerate(parsed_data)
        }

        self.parsed_config.interfaces.update(interfaces_dict)

    def parse_interfaces(self) -> None:
        """Parse physical interfaces."""
        command = "show running physical interfaces"

        parsed_data = parse_output(
            platform=self.platform,
            template_dir=self.custom_templates_dir,
            command=command,
            data=self.running_config,
        )

        for intf in parsed_data:
            intf["enabled"] = map_enabled(intf.get("ENABLED", ""))
            if intf["switchport_mode"] == "trunk":
                if not intf["trunk_allowed_vlans"]:
                    vlans = None
                else:
                    vlans = expand_vlans(intf["trunk_allowed_vlans"])
                intf["trunk_allowed_vlans"] = vlans
            else:
                intf["trunk_allowed_vlans"] = None

            intf["native_vlan"] = (
                int(intf["native_vlan"]) if intf.get("native_vlan") else None
            )
            intf["access_vlan"] = (
                int(intf["access_vlan"]) if intf.get("access_vlan") else None
            )
            intf["voice_vlan"] = (
                int(intf["voice_vlan"]) if intf.get("voice_vlan") else None
            )

            if intf["switchport_mode"]:
                switchport = True
                switchport_mode = intf["switchport_mode"]
            else:
                switchport = False
                switchport_mode = "access"
            intf["switchport"] = switchport
            intf["switchport_mode"] = switchport_mode

            intf["description"] = intf.get("description") or None
            intf["negotiation"] = True if intf.get("negotiation") else False

        interfaces_dict = {
            intf["name"]: self.removekey(
                EthernetCsmacdInterface(index=index, **intf), "metadata"
            )
            for index, intf in enumerate(parsed_data)
        }
        self.parsed_config.interfaces.update(interfaces_dict)

    def parse_lag_interfaces(self) -> None:
        """Parse LAG interfaces."""
        command = "show running lag interfaces"

        parsed_data = parse_output(
            platform=self.platform,
            template_dir=self.custom_templates_dir,
            command=command,
            data=self.running_config,
        )
        for intf in parsed_data:
            intf["enabled"] = map_enabled(intf.get("ENABLED", ""))
            if intf["switchport_mode"] == "trunk":
                if not intf["trunk_allowed_vlans"]:
                    vlans = None
                else:
                    vlans = expand_vlans(intf["trunk_allowed_vlans"])
                intf["trunk_allowed_vlans"] = vlans
            else:
                intf["trunk_allowed_vlans"] = None

            intf["native_vlan"] = (
                int(intf["native_vlan"]) if intf.get("native_vlan") else None
            )
            intf["access_vlan"] = (
                int(intf["access_vlan"]) if intf.get("access_vlan") else None
            )

            if intf["switchport_mode"]:
                switchport = True
                switchport_mode = intf["switchport_mode"]
            else:
                switchport = False
                switchport_mode = "access"
            intf["switchport"] = switchport
            intf["switchport_mode"] = switchport_mode

            intf["description"] = intf.get("description") or None

        interfaces_dict = {
            intf["name"]: self.removekey(
                Ieee8023adLagInterface(index=index, **intf), "metadata"
            )
            for index, intf in enumerate(parsed_data)
        }
        self.parsed_config.interfaces.update(interfaces_dict)

    def parse_system_settings(self) -> None:
        """Parse the system hostname from the configuration content."""
        command = "show running system"

        parsed_data = parse_output(
            platform=self.platform,
            template_dir=self.custom_templates_dir,
            command=command,
            data=self.running_config,
        )

        system_settings = {}

        # hostname
        system_settings["hostname"] = (
            parsed_data[0].get("hostname") if parsed_data[0].get("hostname") else None
        )
        # banner motd
        system_settings["motd_banner"] = (
            " ".join(parsed_data[0].get("banner_motd"))
            if parsed_data[0].get("banner_motd")
            else None
        )

        # banner login
        system_settings["login_banner"] = (
            " ".join(parsed_data[0].get("banner_login"))
            if parsed_data[0].get("banner_login")
            else None
        )

        # domain_name
        system_settings["domain_name"] = (
            parsed_data[0].get("domain_name")
            if parsed_data[0].get("domain_name")
            else None
        )
        # location (from snmp)
        system_settings["location"] = (
            parsed_data[0].get("location") if parsed_data[0].get("location") else None
        )
        # contact (from snmp)
        system_settings["contact"] = (
            parsed_data[0].get("contact") if parsed_data[0].get("contact") else None
        )

        self.parsed_config.system.config = self.removekey(
            SystemConfig(**system_settings), "metadata"
        )

    def parse_ntp(self) -> None:
        """Parse NTP configuration."""
        command = "show running ntp"

        parsed_data = parse_output(
            platform=self.platform,
            template_dir=self.custom_templates_dir,
            command=command,
            data=self.running_config,
        )

        ntp_servers = {}
        for ntp_server in parsed_data:
            if not ntp_server.get("address"):
                continue
            else:
                ntp_server["address"] = str(ntp_server.get("address"))

            ntp_server["version"] = (
                int(ntp_server["version"]) if ntp_server.get("version") else None
            )

            ## Not used in Cisco IOS, but included for completeness
            ##port = entry.get("port")
            ##if port:
            ##    ntp_server["port"] = {"value": int(port)}

            if ntp_server.get("prefer"):
                ntp_server["prefer"] = True
            else:
                ntp_server["prefer"] = False

            if ntp_server.get("source_interface"):
                for intf_name, intf in self.parsed_config.interfaces.items():
                    intf_type = (
                        intf.get("type")
                        if isinstance(intf, dict)
                        else getattr(intf, "type", None)
                    )
                    intf_vlan_id = (
                        intf["vlan_id"].get("value")
                        if isinstance(intf, dict) and intf.get("vlan_id")
                        else getattr(intf, "vlan_id", None)
                    )
                    if intf_type == "l3ipvlan" and intf_vlan_id == int(
                        ntp_server.get("source_interface").replace("Vlan", "")
                    ):
                        intf_ref = ReferenceTo(pointer=f"interfaces.{intf_name}")
                        break
                ntp_server["source_interface"] = intf_ref
            ntp_servers[ntp_server["address"]] = self.removekey(
                NtpServer(**ntp_server), "metadata"
            )

        self.parsed_config.system.ntp.servers = ntp_servers

    def parse_ssh(self) -> None:
        """Parse SSH configuration."""
        command = "show running ssh"

        parsed_data = parse_output(
            platform=self.platform,
            template_dir=self.custom_templates_dir,
            command=command,
            data=self.running_config,
        )

        ssh_values_dict = dict()
        for entry in parsed_data:
            ssh_version = None
            if entry.get("protocol_version"):
                ssh_version = {"value": entry.get("protocol_version")}
                ssh_values_dict["enable"] = {"value": bool(ssh_version)}
                ssh_values_dict["protocol_version"] = ssh_version

            ssh_timeout = None
            if entry.get("timeout"):
                ssh_timeout = {"value": entry.get("timeout")}
                ssh_values_dict["timeout"] = ssh_timeout

            ssh_auth_retries = None
            if entry.get("authentication_retries"):
                ssh_auth_retries = {"value": entry.get("authentication_retries")}
                ssh_values_dict["authentication_retries"] = ssh_auth_retries

            if entry.get("source_interface"):
                for intf_name, intf in self.parsed_config.interfaces.items():
                    intf_type = (
                        intf.get("type")
                        if isinstance(intf, dict)
                        else getattr(intf, "type", None)
                    )
                    intf_vlan_id = (
                        intf["vlan_id"].get("value")
                        if isinstance(intf, dict) and intf.get("vlan_id")
                        else getattr(intf, "vlan_id", None)
                    )
                    if intf_type == "l3ipvlan" and intf_vlan_id == int(
                        entry.get("source_interface").replace("Vlan", "")
                    ):
                        intf_ref = ReferenceTo(pointer=f"interfaces.{intf_name}")
                        break
                ssh_values_dict["source_interface"] = intf_ref

        self.parsed_config.system.ssh.config = self.removekey(
            SshServer(**ssh_values_dict), "metadata"
        )
        self.parsed_config.system.ssh.host_keys = {}

    # Only used for local testing with static config file
    def load_running_config(self, filepath: str) -> str:
        with open(filepath, "r") as f:
            return f.read()

    def parse_snmp_traps(self) -> None:
        """Parse SNMP trap configuration."""
        command = "show running snmp traps"

        parsed_data = parse_output(
            platform=self.platform,
            template_dir=self.custom_templates_dir,
            command=command,
            data=self.running_config,
        )

        snmp_traps_dict = {}
        for entry in parsed_data:
            snmp_trap_values_dict = {}
            # Every trap that is allowed is specified in TrapEvent model, so we loop through all traps and check if they are enabled, if they are enabled we add them to the trap_events list in the SnmpConfig
            # Below code needs to be fixed and handled correctly as the config that Cisco gives does not match exactly how the traps are defined in the model.
            if entry.get("traps"):
                for i, trap in enumerate(entry.get("traps")):
                    snmp_trap_values_dict["name"] = f"trap_{i}"
                    snmp_trap_values_dict["event_name"] = trap
                    snmp_traps_dict[entry.get("event_name")] = self.removekey(
                        TrapEvent(**snmp_trap_values_dict), "metadata"
                    )

        self.parsed_config.system.snmp.trap_events = snmp_traps_dict

    def parse_snmp_views(self) -> None:
        """Parse SNMP view configuration."""
        command = "show running snmp views"

        parsed_data = parse_output(
            platform=self.platform,
            template_dir=self.custom_templates_dir,
            command=command,
            data=self.running_config,
        )

        snmp_views_dict = {}
        for i, entry in enumerate(parsed_data):
            snmp_view_values_dict = {}
            if entry.get("view_name"):
                snmp_view_values_dict["name"] = (
                    f"{entry.get('view_name')}_{i}" if entry.get("view_name") else None
                )
                snmp_view_values_dict["oid"] = (
                    entry.get("view_oid") if entry.get("view_oid") else None
                )
                snmp_view_values_dict["included"] = (
                    True if "included" in entry.get("view_status") else False
                )
                snmp_views_dict[f"{entry.get('view_name')}_{i}"] = self.removekey(
                    SnmpView(**snmp_view_values_dict), "metadata"
                )

        self.parsed_config.system.snmp.views = snmp_views_dict

    def parse_snmp_servers(self) -> None:
        """Parse SNMP server configuration."""
        command = "show running snmp servers"

        parsed_data = parse_output(
            platform=self.platform,
            template_dir=self.custom_templates_dir,
            command=command,
            data=self.running_config,
        )

        snmp_servers_dict = {}
        for entry in parsed_data:
            snmp_server_values_dict = {}
            snmp_server_values_dict["address"] = entry.get("host")
            if entry.get("host"):
                snmp_server_values_dict["address"] = str(entry.get("host"))
                snmp_server_values_dict["enabled"] = True
            else:
                snmp_server_values_dict["address"] = None
                snmp_server_values_dict["enabled"] = False
            snmp_server_values_dict["port"] = (
                int(entry.get("server_port")) if entry.get("server_port") else None
            )
            if entry.get("version"):
                version = (
                    "v3"
                    if entry.get("version") == "3"
                    else "v2c" if entry.get("version") == "2c" else None
                )
                snmp_server_values_dict["version"] = version
            snmp_server_values_dict["community"] = (
                entry.get("community_string") if entry.get("community_string") else None
            )
            snmp_server_values_dict["username"] = (
                entry.get("server_user") if entry.get("server_user") else None
            )
            snmp_server_values_dict["security_level"] = (
                entry.get("server_security_level")
                if entry.get("server_security_level")
                else None
            )
            if entry.get("source_interface"):
                for intf_name, intf in self.parsed_config.interfaces.items():
                    intf_type = (
                        intf.get("type")
                        if isinstance(intf, dict)
                        else getattr(intf, "type", None)
                    )
                    intf_vlan_id = (
                        intf["vlan_id"].get("value")
                        if isinstance(intf, dict) and intf.get("vlan_id")
                        else getattr(intf, "vlan_id", None)
                    )
                    if intf_type == "l3ipvlan" and intf_vlan_id == int(
                        entry.get("source_interface").replace("Vlan", "")
                    ):
                        intf_ref = ReferenceTo(pointer=f"interfaces.{intf_name}")
                        snmp_server_values_dict["source_interface"] = intf_ref
                        break
                    else:
                        snmp_server_values_dict["source_interface"] = None
            snmp_server_values_dict["network_instance"] = (
                entry.get("vrf") if entry.get("vrf") else None
            )

            snmp_servers_dict[entry.get("host")] = self.removekey(
                SnmpServer(**snmp_server_values_dict), "metadata"
            )

        self.parsed_config.system.snmp.trap_servers = snmp_servers_dict

    def parse_snmp_users(self) -> None:
        """Parse SNMP user configuration."""
        command = "show running snmp users"

        parsed_data = parse_output(
            platform=self.platform,
            template_dir=self.custom_templates_dir,
            command=command,
            data=self.running_config,
        )

        snmp_users_dict = {}
        for i, entry in enumerate(parsed_data):
            snmp_user_values_dict = {}
            snmp_user_values_dict["name"] = f"user_{i}"
            snmp_user_values_dict["username"] = (
                entry.get("user") if entry.get("user") else None
            )
            # Security levels
            #   auth    group using the authNoPriv Security Level
            #   noauth  group using the noAuthNoPriv Security Level
            #   priv    group using SNMPv3 authPriv security level
            if entry.get("security_level") == "auth":
                snmp_user_values_dict["security_level"] = "AUTH_NO_PRIV"
            elif entry.get("security_level") == "noauth":
                snmp_user_values_dict["security_level"] = "NO_AUTH_NO_PRIV"
            elif entry.get("security_level") == "priv":
                snmp_user_values_dict["security_level"] = "AUTH_PRIV"
            else:
                None
            snmp_user_values_dict["auth_protocol"] = (
                entry.get("auth_protocol") if entry.get("auth_protocol") else None
            )
            snmp_user_values_dict["auth_password"] = (
                entry.get("auth_password") if entry.get("auth_password") else None
            )
            snmp_user_values_dict["priv_protocol"] = (
                entry.get("priv_protocol") if entry.get("priv_protocol") else None
            )
            snmp_user_values_dict["priv_password"] = (
                entry.get("priv_password") if entry.get("priv_password") else None
            )

            snmp_users_dict[f"user_{i}"] = self.removekey(
                SnmpUser(**snmp_user_values_dict), "metadata"
            )

        self.parsed_config.system.snmp.users = snmp_users_dict

    def parse_snmp_communities(self) -> None:
        """Parse SNMP community configuration."""
        command = "show running snmp communities"

        parsed_data = parse_output(
            platform=self.platform,
            template_dir=self.custom_templates_dir,
            command=command,
            data=self.running_config,
        )

        snmp_communities_dict = {}
        for i, entry in enumerate(parsed_data):
            snmp_community_values_dict = {}
            snmp_community_values_dict["name"] = f"community_{i}"
            snmp_community_values_dict["community"] = (
                entry.get("community_string") if entry.get("community_string") else None
            )
            snmp_community_values_dict["access"] = (
                entry.get("access") if entry.get("access") else None
            )
            # Should be a reference to an existing SnmpView
            # snmp_community_values_dict['view'] = entry.get("view") if entry.get("view") else None
            snmp_community_values_dict["ipv4_acl"] = (
                entry.get("ipv4_acl") if entry.get("ipv4_acl") else None
            )
            snmp_community_values_dict["ipv6_acl"] = (
                entry.get("ipv6_acl") if entry.get("ipv6_acl") else None
            )
            if entry.get("source_interface"):
                for intf_name, intf in self.parsed_config.interfaces.items():
                    intf_type = (
                        intf.get("type")
                        if isinstance(intf, dict)
                        else getattr(intf, "type", None)
                    )
                    intf_vlan_id = (
                        intf["vlan_id"].get("value")
                        if isinstance(intf, dict) and intf.get("vlan_id")
                        else getattr(intf, "vlan_id", None)
                    )
                    if intf_type == "l3ipvlan" and intf_vlan_id == int(
                        entry.get("source_interface").replace("Vlan", "")
                    ):
                        intf_ref = ReferenceTo(pointer=f"interfaces.{intf_name}")
                        break
                snmp_community_values_dict["source_interface"] = intf_ref

            snmp_communities_dict[f"community_{i}"] = self.removekey(
                SnmpCommunity(**snmp_community_values_dict), "metadata"
            )
        self.parsed_config.system.snmp.communities = snmp_communities_dict

    def parse_snmp(self) -> None:
        """Parse SNMP configuration."""
        command = "show running snmp"

        parsed_data = parse_output(
            platform=self.platform,
            template_dir=self.custom_templates_dir,
            command=command,
            data=self.running_config,
            # data=self.load_running_config("/Users/jani/scripts/acex/docs/examples/test_run/sample_running.txt") # Using this for testing with a static config file, replace with self.running_config for actual use
        )

        # SNMP parsing logic would go here, similar to NTP and SSH parsing
        snmp_config = {}
        snmp_config_values_dict = {}
        for entry in parsed_data:
            if entry.get("host"):
                snmp_config_values_dict["enabled"] = True
            snmp_config_values_dict["location"] = (
                entry.get("location") if entry.get("location") else None
            )
            snmp_config_values_dict["contact"] = (
                entry.get("contact") if entry.get("contact") else None
            )
            snmp_config_values_dict["engine_id"] = (
                entry.get("engine_id") if entry.get("engine_id") else None
            )

            snmp_config = self.removekey(
                SnmpConfig(**snmp_config_values_dict), "metadata"
            )

        self.parsed_config.system.snmp.config = snmp_config
